import 'package:fis_measure/interfaces/process/calculators/values.dart';
import 'package:fis_measure/interfaces/process/items/item.dart';
import 'package:fis_measure/interfaces/process/items/item_feature.dart';
import 'package:fis_measure/process/unit/convert/convert.dart';
import 'package:fis_measure/values/unit_desc.dart';

class FeatureValueDescConverter {
  final IMeasureItemFeature feature;
  FeatureValueDescConverter(this.feature);

  String idPlaceStr = ' ';
  String idStr = ' ';

  List<String> generate(int idLength) {
    idPlaceStr = ' '.padRight(idLength, ' ');
    idStr = feature.id.toString().padRight(idLength, ' ');

    final List<String> arr = [];
    final mainLines = _findFeatureLines(feature);
    arr.addAll(mainLines);

    final ref = feature.refItem;
    if (ref is ITopMeasureItem) {
      for (var child in ref.childItems) {
        for (var feature in child.measuredFeatures) {
          final str = _findChildLine(feature);
          if (str != null && str.isNotEmpty) {
            arr.add(str);
          }
        }
        if (child.feature != null) {
          final str = _findChildLine(child.feature!);
          if (str != null && str.isNotEmpty) {
            arr.add(str);
          }
        }
      }
    }

    return arr;
  }

  String? _findChildLine(IMeasureItemFeature feature) {
    if (feature.value == null) return null;

    final value = feature.value!;
    final valueStr = _pickValueStr(value);

    final name = feature.refItem.description;

    return ' $idPlaceStr $name    $valueStr';
  }

  List<String> _findFeatureLines(IMeasureItemFeature feature) {
    final count = feature.values.length;
    final itemName = feature.refItem.displayName;
    if (count == 0) {
      return ['$idStr $itemName'];
    }

    List<String> arr = [];
    for (var i = 0; i < count; i++) {
      final val = feature.values[i];
      final meta = val.meta;
      final valueStr = _pickValueStr(val);

      if (i == 0) {
        arr.add('$idStr $itemName    $valueStr');
      } else {
        final name =
            _findDisplayName(meta.description, meta.briefAnnotation, false);
        arr.add('$idPlaceStr $name    $valueStr');
      }
    }

    return arr;
  }

  String _pickValueStr(ValueBase value) {
    if (value is FloatValue) {
      final floatVal = _roundDouble(value.value!, value.meta.fractionalDigits);
      if (floatVal.isNaN) {
        return '';
      }

      final unit = value.unit;
      final targetUnit = value.targetUnit;
      final targetVal = UnitValueConverter.convert(unit, targetUnit, floatVal);
      final unitStr = UnitDescriptionMap.getDesc(targetUnit);
      return '$targetVal $unitStr';
    } else if (value is StringValue) {
      return value.value!;
    }
    return '';
  }

  String _findDisplayName(String description, String? briefAnnotation,
      [bool showAnnotation = false]) {
    if (showAnnotation) {
      if (briefAnnotation != null && briefAnnotation.isNotEmpty) {
        return briefAnnotation;
      }
    }
    return description;
  }

  double _roundDouble(double value, [int digits = 2]) {
    final digitsStr = value.toStringAsFixed(digits);
    final result = double.parse(digitsStr);
    return result;
  }
}
